<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Criteria\DataSet;

// Import framework stuff
use Org\Mxchange\CoreFramework\Criteria\BaseCriteria;
use Org\Mxchange\CoreFramework\Criteria\Local\LocalSearchCriteria;
use Org\Mxchange\CoreFramework\Criteria\Storing\StoreableCriteria;
use Org\Mxchange\CoreFramework\EntryPoint\ApplicationEntryPoint;
use Org\Mxchange\CoreFramework\Generic\FrameworkInterface;

// Import SPL stuff
use \BadMethodCallException;
use \InvalidArgumentException;

/**
 * A set of data storeable in databases
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2023 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class DataSetCriteria extends BaseCriteria implements StoreableCriteria {
	/**
	 * Table name
	 */
	private $tableName = '';

	/**
	 * Unique key
	 */
	private $uniqueKey = '';

	/**
	 * Primary key
	 */
	private $primaryKey = '';

	/**
	 * Primary keys
	 */
	private $primaryKeys = [];

	/**
	 * Search criteria instance
	 */
	private $searchInstance = NULL;

	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	private function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Creates an instance of this criteria
	 *
	 * @param	$tableName			Name of the table
	 * @return	$criteriaInstance	An instance of this criteria
	 * @throws	InvalidArgumentException	If a parameter is not valid
	 */
	public static final function createDataSetCriteria (string $tableName) {
		// Validate parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('DATA-SET-CRITERIA: tableName=%s - CALLED!', $tableName));
		if (empty($tableName)) {
			// Throw IAE
			throw new InvalidArgumentException('Parameter "tableName" is empty', FrameworkInterface::EXCEPTION_INVALID_ARGUMENT);
		}

		// Get a new instance
		$criteriaInstance = new DataSetCriteria();

		// Set table name
		$criteriaInstance->setTableName($tableName);

		// Return the instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('DATA-SET-CRITERIA: criteriaInstance=%s - EXIT!', $criteriaInstance->__toString()));
		return $criteriaInstance;
	}

	/**
	 * Setter for table name
	 *
	 * @param	$tableName	Name of the table to set
	 * @return	void
	 */
	public final function setTableName (string $tableName) {
		$this->tableName = $tableName;
	}

	/**
	 * Getter for table name
	 *
	 * @return	$tableName	Name of the table to set
	 */
	public final function getTableName () {
		return $this->tableName;
	}

	/**
	 * Setter for unique key
	 *
	 * @param	$uniqueKey	Column to use as unique key
	 * @return	void
	 */
	public final function setUniqueKey (string $uniqueKey) {
		$this->uniqueKey = $uniqueKey;
	}

	/**
	 * Getter for unique key
	 *
	 * @return	$uniqueKey	Column to use as unique key
	 */
	public final function getUniqueKey () {
		return $this->uniqueKey;
	}

	/**
	 * Setter for primary key
	 *
	 * @param	$primaryKey		Primary key to set
	 * @return	void
	 */
	public final function setPrimaryKey (string $primaryKey) {
		$this->primaryKey = $primaryKey;
	}

	/**
	 * Setter for primary key array
	 *
	 * @param	$primaryKeys	Primary key array to set
	 * @return	void
	 */
	public function setPrimaryKeyCombined (array $primaryKeys) {
		$this->primaryKeys = $primaryKeys;
	}

	/**
	 * Getter for primary keys
	 *
	 * @return	$primaryKeys	Primary key array
	 */
	public final function getPrimaryKeys () {
		// Return it
		return $this->primaryKeys;
	}

	/**
	 * Setter for search instance
	 *
	 * @param	$searchInstance		Searchable criteria instance
	 * @return	void
	 */
	public final function setSearchInstance (LocalSearchCriteria $searchInstance) {
		$this->searchInstance = $searchInstance;
	}

	/**
	 * Getter for search instance
	 *
	 * @return	$searchInstance		Searchable criteria instance
	 */
	public final function getSearchInstance () {
		return $this->searchInstance;
	}

	/**
	 * Getter for unique key value
	 *
	 * @return	$uniqueValue	Value of the unique key
	 * @throws	BadMethodCallException	If no primary and no unique key was found
	 */
	public final function getUniqueValue () {
		// Get primary key(s) first
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('DATA-SET-CRITERIA: CALLED!');
		$primaryKey  = trim($this->getCriteriaElemnent($this->getPrimaryKey()));
		$primaryKeys = $this->getPrimaryKeys();

		/*
		 * If this is not set, this could mean a badly written frontend as
		 * tables should always have a primary key.
		 */
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('DATA-SET-CRITERIA: this->tableName=%s,primaryKey=%s,primaryKeys()=%d', $this->getTableName(), $primaryKey, count($primaryKeys)));
		if (count($primaryKeys) > 0) {
			/*
			 * Init return value, this can be put all together without any
			 * separator as it only aids as a "unique value" for generating the
			 * "row file name".
			 */
			$return = '';

			// Combination set, so get all
			foreach ($primaryKeys as $primaryKeyPart) {
				// Add it
				/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('DATA-SET-CRITERIA: primaryKeyPart=%s', $primaryKeyPart));
				$return .= trim($this->getCriteriaElemnent($primaryKeyPart));
			}

			// Debug message

			// Return it
			/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('DATA-SET-CRITERIA: return=%s - EXIT!', $return));
			return $return;
		} elseif (!empty($primaryKey)) {
			// Return primary key
			/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('DATA-SET-CRITERIA: primaryKey=%s - EXIT!', $primaryKey));
			return $primaryKey;
		} else {
			// Issue a warning
			self::createDebugInstance(__CLASS__, __LINE__)->warningMessage('DATA-SET-CRITERIA: Primary key not set for table ' . $this->getTableName() . ', please fix your table. Falling back to unique key ...');

			// Get unique key
			$uniqueKey = trim($this->getCriteriaElemnent($this->getUniqueKey()));

			// Is it empty, too?
			/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('DATA-SET-CRITERIA: uniqueKey=%s', $uniqueKey));
			if (empty($uniqueKey)) {
				// Bad news, nothing is "unique" by design for this table
				throw new BadMethodCallException(sprintf('Table %s has both no primary and unique key, but %s was called. Please fix your table.', $this->getTableName(), __METHOD__), FrameworkInterface::EXCEPTION_BAD_METHOD_CALL);
			}

			// Return unique key
			/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('DATA-SET-CRITERIA: uniqueKey=%s - EXIT!', $uniqueKey));
			return $uniqueKey;
		}
	}

	/**
	 * Getter for primary key or unique key if not set
	 *
	 * @return	$primaryKey		Primary key or unique key if not set
	 */
	public final function getPrimaryKey () {
		// Get primary key by default
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('DATA-SET-CRITERIA: CALLED!');
		$primaryKey = $this->primaryKey;

		// Is it still empty?
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('DATA-SET-CRITERIA: primaryKey=%s', $primaryKey));
		if (empty($primaryKey)) {
			// Get uniqueKey
			$primaryKey = $this->getUniqueKey();
		}

		// Return it
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('DATA-SET-CRITERIA: primaryKey=%s - EXIT!', $primaryKey));
		return $primaryKey;
	}

}
