<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Index\File\Stack;

// Import framework stuff
use Org\Mxchange\CoreFramework\Generic\FrameworkInterface;
use Org\Mxchange\CoreFramework\Index\File\BaseFileIndex;
use Org\Mxchange\CoreFramework\Index\File\Stack\IndexableStack;
use Org\Mxchange\CoreFramework\Index\Indexable;
use Org\Mxchange\CoreFramework\Registry\Registerable;
use Org\Mxchange\CoreFramework\Stack\File\StackableFile;

// Import SPL stuff
use \InvalidArgumentException;
use \SplFileInfo;
use \UnexpectedValueException;

/**
 * A FileStack index class
 *
 * @author		Roland Haeder <webmaster@ship-simu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2023 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.ship-simu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class FileStackIndex extends BaseFileIndex implements IndexableStack, Registerable {
	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	private function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Creates an instance of this Index class and prepares it for usage
	 *
	 * @param	$fileInfoInstance	An instance of a SplFileInfo class
	 * @return	$indexInstance	An instance of this Index class
	 */
	public final static function createFileStackIndex (SplFileInfo $fileInfoInstance) {
		// Get a new instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FILE-STACK-INDEX: fileInfoInstance[%s]=%s - CALLED!', get_class($fileInfoInstance), $fileInfoInstance));
		$indexInstance = new FileStackIndex();

		// Initialize index
		$indexInstance->initFileIndex($fileInfoInstance);

		// Return the prepared instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FILE-STACK-INDEX: indexInstance=%s - EXIT!', $indexInstance->__toString()));
		return $indexInstance;
	}

	/**
	 * Adds given data's hash to an index file
	 *
	 * @param	$stackName	Name of stack to add hash for
	 * @param	$data		Hash and gap position to be added to the index
	 * @return	void
	 * @throws	InvalidArgumentException	If a parameter is not valid
	 * @throws	UnexpectedValueException	If an invalid gap position is being returned
	 */
	public function addHashedDataToIndex (string $stackName, array $data) {
		// Validate parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FILE-STACK-INDEX: stackName=%s,data()=%d - CALLED!', $stackName, count($data)));
		if (empty($stackName)) {
			// Throw IAE
			throw new InvalidArgumentException('Parameter "stackName" is empty', FrameworkInterface::EXCEPTION_INVALID_ARGUMENT);
		} elseif (count($data) == 0) {
			// Throw it again
			throw new InvalidArgumentException('Parameter "data" is an empty array');
		} elseif (!isset($data[StackableFile::ARRAY_NAME_HASH])) {
			// Important array element missing
			throw new InvalidArgumentException(sprintf('data[%s] not found', StackableFile::ARRAY_NAME_HASH));
		} elseif (!isset($data[StackableFile::ARRAY_NAME_GAP_POSITION])) {
			// Important array element missing
			throw new InvalidArgumentException(sprintf('data[%s] not found', StackableFile::ARRAY_NAME_GAP_POSITION));
		} elseif (!isset($data[StackableFile::ARRAY_NAME_DATA_LENGTH])) {
			// Important array element missing
			throw new InvalidArgumentException(sprintf('data[%s] not found', StackableFile::ARRAY_NAME_DATA_LENGTH));
		}

		// Raw data been written to the file
		$rawData = sprintf('%s%s%s%s%s%s%s',
			$stackName,
			Indexable::SEPARATOR_GROUP_HASH,
			$data[StackableFile::ARRAY_NAME_HASH],
			Indexable::SEPARATOR_HASH_GAP_POSITION,
			$data[StackableFile::ARRAY_NAME_GAP_POSITION],
			Indexable::SEPARATOR_GAP_LENGTH,
			$data[StackableFile::ARRAY_NAME_DATA_LENGTH]
		);
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FILE-STACK-INDEX: stackName=%s,hash{}=0x%s,rawData(%d)=%s', $stackName, bin2hex($data[StackableFile::ARRAY_NAME_HASH]), strlen($rawData), $rawData));

		// Search for next free gap
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FILE-STACK-INDEX: Invoking this->iteratorInstance->binaryFileInstance->searchNextGap(%d) ...', strlen($rawData)));
		$gapPosition = $this->getIteratorInstance()->getBinaryFileInstance()->searchNextGap(strlen($rawData));

		// Gap position cannot be smaller or equal than header length
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FILE-STACK-INDEX: stackName=%s,hash{}=0x%s,gapPosition=%s', $stackName, bin2hex($data[StackableFile::ARRAY_NAME_HASH]), $gapPosition));
		if ($gapPosition <= ($this->getIteratorInstance()->getBinaryFileInstance()->getHeaderSize() + 1)) {
			// Not valid gap position returned
			throw new UnexpectedValueException(sprintf('gapPosition[%s]=%d is smaller or equal headerSize+1=%d', gettype($gapPosition), $gapPosition, ($this->getIteratorInstance()->getBinaryFileInstance()->getHeaderSize() + 1)));
		}

		// Then write the data at that gap
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FILE-STACK-INDEX: Invoking this->iteratorInstance->binaryFileInstance->writeData(%d,%s) ...', $gapPosition, $rawData));
		$this->getIteratorInstance()->getBinaryFileInstance()->writeData($gapPosition, $rawData);

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FILE-STACK-INDEX: stackName=%s,hash{}=0x%s,rawData()=%d - EXIT!', $stackName, bin2hex($data[StackableFile::ARRAY_NAME_HASH]), strlen($rawData)));
	}

}
